/**
 * @file Fonctions pour interagir avec la BDD sigma et le LDAP.
 * @author akka vodol
 */
import knex from '../../db/knex_router';
import { renseignerSurUtilisateur, repliquerTOLdesIds, listerGroupes, listerMembres, listerAdministrateurs } from '../ldap/ldap_data';
import { exportAllDeclaration } from 'babel-types';

export { renseignerSurUtilisateur, repliquerTOLdesIds, listerMembres };

/**
 * @summary Génère une promise.
 * @function
 * @desc Les fonctions ici sont toutes supposées renvoyer une promise. 
 * Si on veut renvoyer une valeur directement, cette fonction permet de construire une promise qui renvoie cette valeur facilement.
 * @arg {Object} val - La valeur qu'on veut renvoyer.
 * @return {Promise(Object)} Une promise qui renvoi val
 */
const quickPromise = (val) => {
    return new Promise( (resolve, reject) => {
        resolve(val);
    });
};

/**
 * @summary Renvoie le type d'un groupe.
 * @function
 * @desc Parcours les BDD pour savoir dans laquelle se trouve le groupe ayant l'UID donné. 
 * Cette opération nécéssite un parcours de base de donnée, et il est préférable de ne pas sans servir si on a un autre moyend de connaitre le typed d'un groupe
 * (Par exemple, si on dispose d'un retour de BDD pour ce groupe. L'existence du champ School indique alors son type.)
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @arg {Object} groupUID - L'id du groupe dont on veut connaître le type.
 * @return {Promise(String)} Un string représentant le type du groupe. Peut être "SimpleGroup" ou "MetaGroup". Renvoie *Undefined* si le groupe 'existe pas
 */
export const getGroupType = (user, groupUID) => {
    return knex('simple_groups').select('uid').where('uid', groupUID).then( sg_res => {
        if(sg_res)
            return "SimpleGroup";
        return knex('meta_groups').select('uid').where('uid', groupUID).then(mg_res => {
            if(mg_res)
                return "MetaGroup";
            return undefined;
        });
    });
};

/**
 * @summary Renvoie tous les utilisateurs ayant des droits d'administrateur sur un groupe.
 * @function
 * @desc Les utilisateurs qui ont un droit d'administrateur sur un groupe simple sont ses administrateurs 
 * et les utilisateurs ayant droit d'admin sur son parent.
 * Les utilisateurs ayant droit d'admin sur un meta-groupe sont les utilisateurs 
 * ayant droit d'admin sur un des groupes membres
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste de tous les utilisateurs ayant droit d'admin sur le groupe
 */
export const getUsersWithAdminRights = (user, groupUID) => {
    return getGroupType(user, groupUID).then( groupType => {
        
        if(groupType == "SimpleGroup"){
            return listerAdministrateurs(user, groupUID).then(adminList => {
                if (typeof adminList == "undefined")
                    return undefined;
                else
                    return knex('simple_groups').select('parentuid').where('uid', groupUID).then(req => {
                        if (req[0].parentuid)
                            return getUsersWithAdminRights(user, req[0].parentuid).then(parentAdmins => {
                                return adminList.concat(parentAdmins);
                            });
                        else
                            return adminList.concat(['anatole.romon']); // pour les besoins des tests, anatole romon a tout les droits
                    });
            });
        }else{
            return getMetaGroupAdminMembers(user, groupUID).then(mg_members => {
                function fuse (l){
                    if(l)
                        return getUsersWithAdminRights(l.pop()).then(usrs => fuse(l).concat(usrs) );
                    else
                        return [];
                }
            });
        }
    });
};

/**
 * @summary teste si un utilisateur a des droits
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 */
export const hasAdminRights = (user, groupUID) => {
    if(user.uid == "anatole.romon")
        return quickPromise(true);
    return getUsersWithAdminRights(user, groupUID).then(adminList => {
        return (typeof adminList != "undefined" && adminList.indexOf(user.uid) != -1);
    });
};


/**
 * @summary Renvoie une liste des id de tous les groupes visibles par l'utilisateur
 * @desc Cette fonction génère un callback qui créé une table contenant les uid de tous les groupes visibles
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise(Callback)} callback contruisant une requête knex pour une table de tous les id visibles.
 */
export const getVisibleGroupCallback = (user) => {
    return listerGroupes(user, user.uid).then(group_ids => {
        return function (global_query_builder){
            if (typeof group_ids == "undefined")
                throw "invalid user";
            var membered_groups = qb => qb.select('simple_groups.uid').from('simple_groups').whereIn('uid', group_ids.concat(['kes']));
            var directly_visible_simple_groups = qb =>  qb.with('membered_groups', membered_groups).select('simple_groups.uid').from('simple_groups').distinct()
                .innerJoin('membered_groups',
                    function () {
                        this.on('simple_groups.uid', '=', 'membered_groups.uid').orOn('simple_groups.parentuid', '=', 'membered_groups.uid');
                    }
                );
            return directly_visible_simple_groups(global_query_builder);
        }
    });
    
};

/**
 * @summary Renvoie un unique groupe, ssi ce groupe est visible par l'utilisateur
 * @desc Actuellement, la fonction effectue la même requête que *getAllVisibleGroups* et restreint au groupe demandé. Cette fonction peut être implémentée de manière plus efficace et plus chiante.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @arg {String} uid - uid du groupe que l'on veut voire. 
 * @return {Promise(group)} Retour de requête knex. Le groupe demandé, si l'utilisateur a le droit de la voire.
 */
export const getGroupIfVisible = (user, groupUID) => {
    return getVisibleGroupCallback(user).then(visible_groups => {
        return knex.with('visible_groups', visible_groups).select('groups.*').from('groups').innerJoin('visible_groups', function (){
            this.on('visible_groups.uid', '=', 'groups.uid');
        }).where('groups.uid', groupUID).then(res => {
            return res[0];
        });
    });
};

/**
 * @summary Renvoie tous les groupes visibles par l'utilisateur user
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 */
export const getAllVisibleGroups = (user) => {
    return getVisibleGroupCallback(user).then(visible_groups => {
        return knex.with('visible_groups', visible_groups).select('groups.*').from('groups').innerJoin('visible_groups', function (){
            this.on('visible_groups.uid', '=', 'groups.uid');
        });
    });
};

/**
 * @summary Teste si un utilisateur est membre d'un groupe
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @arg {Object} groupUID - L'id du groupe dont on veu savoir si l'utilisateur est membre. 
 * @return {Promise(Boolean)} Boolean indiquant si l'utilisateur est membre du groupe.
 */
export const isMember = (user, groupUID) => {
    return listerGroupes(user, user.uid).then(group_ids => group_ids && group_ids.indexOf(groupUID) != -1);
};

/**
 * @summary Attribue un UID qui n'a pas encore été utilisé à un groupe
 * @desc RASifie le string initialUID si necessaire (ramené à de l'ASCCI sans espace), puis si l'uid est deja pris rajoute un n a la fin et reteste
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste de tous les utilisateurs ayant droit d'admin sur le groupe
 */
export const getAvailablegroupUID = (initialUID) => {
    let rasUID = initialUID.replace(' ', '_').replace(/\W/g, '').toLowerCase();
    return knex.from('groups').where('uid', rasUID).then(res => {
        if (res.length == 0) {
            return (rasUID);
        } else {
            return (getAvailablegroupUID(rasUID + 'n'));
        }
    });
};

/**
 * @summary Créé un groupe si les arguments sont tous valides
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * Les authorisations de l'utilisateur ne sont pas vérifiées
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction *getUsersWithAdminRights*
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 */
export const createSubgroup = (user, args) => {
    if (typeof args.parentuid != 'string')
        throw "Illegal argument : parentuid must be a non null string";
    if (typeof args.name != 'string')
        throw "Illegal argument : name must be a non null string";

    return (getAvailablegroupUID(args.uid).then(rasUID => {
        // TODO : appeller une fonction de ldap_data pour y créer un groupe.
        return knex('simple_groups').insert({
            uid: rasUID,
            parentuid: args.parentuid,
            createdAt: knex.fn.now(),
            updatedAt: this.createdAt,
            name: args.name,
            website: args.website,
            description: args.description,
            school: args.school
        }).then(res => {
            return getGroupIfVisible(user, rasUID);
        });
    }));
};

/**
 * @summary Créé un groupe si les arguments sont tous valides et l'utilisateur est authorisé
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction *getUsersWithAdminRights*
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 */
export const createGroupIfLegal = (user, args) => {
    return hasAdminRights(user, args.parentuid).then(answer => {
        if (!answer)
            throw "illegal request : you must have admin rights over a group to create a subgroup of that group";
        return createSubgroup(user, args);
    });
};

export const getUserJoinGroupRequests = (user, recipientUID) => {
    return knex('user_join_group').select(/*'id', 'useruid', 'message'*/)/*.where('recipient', recipientUID)*/;
};

//Don't forget the argument user is the guy who makes the request, not the user we want
export const getUser = (user, userUID) => { 
    return new Promise( (resolve, reject) => {
        resolve({
            givenName : "Jean-Michel",
            lastName : "Samarchepas",
            uid : "jm",
            birthdate : "dans le turfu",
            mail : "email.email@email.com",
            phone : "06 60 06 60 06",
            groups : [],
            likes : [],
            address : "aussi dans le turfu"
        });
    });
    /*return renseignerSurUtilisateur(user, userUID).then(res => {
        return {
            givenName : res[0].givenName,
            lastName : res[0].lastName,
            uid : userUID,
            birthdate : res[0].brBirthdate,
            mail : res[0].mail,
            phone : res[0].telephoneNumber,
            groups : [],
            likes : [],
            address : res[0].brRoom
        };
    });*/
};

export const getEvent = (user, eventID) => {
    return quickPromise(null);
};

export const getGroup = (user, groupUID) => {
    return quickPromise(null);
};

export const getSimpleGroup = (user, groupUID) => {
    return quickPromise(null);
};

export const getMetaGroup = (user, groupUID) => {
    return quickPromise(null);
};

export const getMetaGroupAdminMembers = (user, metaGroupUID) => {
    return quickPromise([]);
};